
#include <mpi.h>
#include <cstdio>
#include <cstdlib>
#include <cmath>

#define NMAX 100
#define NPMAX 10

#define L 10                    // количество уравнений и неизвестных

#define comm MPI_COMM_WORLD

void Split(double*A,double*B,double*C,double*D,int m1,int m2,double*X) {	// Метод разбиений

int IP,NP,N,m,i;
double W[2][4],ww[100][4],alpha[100],beta[100],XX[2];

  MPI_Comm LComm;
  MPI_Comm_dup(MPI_COMM_WORLD,&LComm);

  MPI_Comm_size( LComm, &NP ); //NP - количество процессов
  MPI_Comm_rank( LComm, &IP ); //IP - номер каждого процесса
	
// Прогонка матрицы вниз
    for (m=m1+1; m<m2; m++) {
      A[m]=A[m]/B[m]; C[m]=C[m]/B[m]; D[m]=D[m]/B[m];
      B[m+1]=B[m+1]-C[m]*A[m+1]; D[m+1]=D[m+1]-D[m]*A[m+1]; A[m+1]= -A[m]*A[m+1];
    }

// Прогонка матрицы вверх
    for (m=m2-1; m>m1+1; m--) {
      A[m-1]=A[m-1]-A[m]*C[m-1]; D[m-1]=D[m-1]-D[m]*C[m-1]; C[m-1]= -C[m]*C[m-1];
    }
    B[m1]=B[m1]-A[m1+1]*C[m1]; D[m1]=D[m1]-D[m1+1]*C[m1]; C[m1]= -C[m1+1]*C[m1];

    W[0][0]=A[m1]; W[0][1]=B[m1]; W[0][2]=C[m1]; W[0][3]=D[m1];
    W[1][0]=A[m2]; W[1][1]=B[m2]; W[1][2]=C[m2]; W[1][3]=D[m2];

    MPI_Gather(&W[0][0],8,MPI_DOUBLE,&ww[0][0],8,MPI_DOUBLE,0,LComm);

    if (IP==0) {

// Решение малой системы методом прогонки

      alpha[1]= - ww[0][2]/ww[0][1]; beta[1]=ww[0][3]/ww[0][1];
      for (i=1; i<2*NP-1; i++) {
        alpha[i+1]= - ww[i][2]/(ww[i][0]*alpha[i]+ww[i][1]);
        beta[i+1]= - (ww[i][0]*beta[i]-ww[i][3])/(ww[i][0]*alpha[i]+ww[i][1]);
      }
      X[2*NP-1]=-(ww[2*NP-1][0]*beta[2*NP-1]-ww[2*NP-1][3])/(ww[2*NP-1][0]*alpha[2*NP-1]+ww[2*NP-1][1]);
      for (i=2*NP-1; i>0; i--) {
        X[i-1]=alpha[i]*X[i]+beta[i];
      }
    } // конец решения малой системы

    MPI_Scatter(&X[0],2,MPI_DOUBLE,&XX[0],2,MPI_DOUBLE,0,LComm);

// Получение решений

    X[m1]=XX[0]; X[m2]=XX[1];
    for (m=m1+1; m<m2; m++) {
     X[m]=D[m]-A[m]*X[m1]-C[m]*X[m2];
    }
// конец метода разбиений

    MPI_Comm_free(&LComm);
    return;
}

int main() {

  FILE *fp;
  int IP,NP,N,NN[NPMAX],m1,m2,n,m,i,m0,mm0[NPMAX];
  double a[NMAX+2],b[NMAX+2],c[NMAX+2],d[NMAX+2],x[NMAX+2];
  float AA,BB,CC,DD;
  
// Режим MPI. Разметка процессоров
  
  MPI_Status status;

  int argc = 0; char **argv = NULL;
  MPI_Init(&argc, &argv);
  MPI_Comm_size( comm, &NP ); //NP - количество процессов
  MPI_Comm_rank( comm, &IP ); //IP - номер каждого процесса

  for(i=0; i<NP; i++) {
    NN[i]=L/NP;                   //NN[i] - количество уравнений у i - го процесса
    if (i<(L%NP)) NN[i]=NN[i]+1;
  }
  mm0[0]=0;
  for (i=1; i<NP; i++) {
    mm0[i]=mm0[i-1]+NN[i-1];
  }
  m0=mm0[IP];
  N=NN[IP];              //N - количество уравнений у каждого процесса

  m1=1; //m1 - номер первого уравнения в каждом процессе
  m2=N; //m2 - номер последнего уравнения в каждом процессе

  printf(" IP=%2d  NP=%2d    N=%4d\n",IP,NP,N);

//-----------------------------------------------------------------------------------------

//заполнение A, B, C, D

fp= fopen("test.txt", "r");
	for(int i=0;i<m0;i++)
		fscanf(fp," %f %f %f %f",&AA,&BB,&CC,&DD);
	for(int i=m1;i<=m2;i++) {
		fscanf(fp," %f %f %f %f",&AA,&BB,&CC,&DD);
		a[i]=AA; b[i]=BB; c[i]=CC; d[i]=DD;
		printf("IP=%d %10.3f %10.3f %10.3f %10.3f\n",IP,a[i],b[i],c[i],d[i]);
	}
fclose(fp);

Split(a,b,c,d,m1,m2,x);


// Печать результатов

	for(i=m1;i<=m2;i++) 
		printf(" x[%d]=%10.3f\n",i+m0,x[i]);

// Конец MPI

        MPI_Finalize();

	return 0;
}